#ifndef LDL_LOGISTIC_H
#define LDL_LOGISTIC_H
#include <vector>
#include <string>
#include <iostream>
#include <eigen3/Eigen/Dense>
#include <functional> //std::function
#include "common.h"
using Eigen::MatrixXd;
using Eigen::VectorXd;
using namespace common;

//for parameter estimation for binary phenotype
//see perform_EM() method
namespace ldl_logistic
{

    double calc_d_val_for_quadrature(                                
                              double Y_val,
                              double exp_val );


    void calc_expectation(    double gamma,
                              double sqrt_2_a_val,
                              double Y_val, double X_beta_dot_val,
                              const VectorXd & ksi_vec,
                              const VectorXd & Z_vec,
                              const measurement & measure,
         const std::vector <std::function<double(double,double)> > & fp_vec,
                              std::vector<double> & expect_vec );
                             

    
    double func_g1( double S_val, 
                    double exp_val 
                    );

    double func_g2( double S_val, 
                     double exp_val
                    ); 


    double func_g3( double S_val,    
                    double exp_val
                    );

    double func_g4( double S_val,
                    double exp_val
                   );


    double func_g5( double S_val,
                    double exp_val
                    );

    double func_g6( double S_val,
                    double exp_val
                    );  

    double func_g7( double S_val,
                    double exp_val
                    );  

    double func_g8( double S_val,
                    double exp_val
                    );
  
    double func_g9( double S_val,
                    double exp_val
                    );  
    double func_g10( double S_val,
                    double exp_val
                    );

    double func_g11( double S_val,
                    double exp_val
                    );

    double func_g12( double S_val,
                    double exp_val
                    );

    double func_g13( double S_val,
                    double exp_val
                    );

    double func_g14( double S_val,
                    double exp_val
                    );
    
    //perform_EM()
    //performs parameter estimation using an EM 
    //algorithm.  
    //return value: error code as int.  NO_ERROR if no error.
    //NOT_CONVERGED_ERROR if algorithm didd not converge.
    //pre: Y_vec is a binary phenotype (0.0 or 1.0). 
    //Y_vec.rows() == number of subjects (no missing phenotype).
    //X_mat == matrix of covariates (including constant) 
    //X_mat.rows() == number of covariates (including constant)
    //The SNP may or may not be included as a row in X_mat.
    //X_mat.cols() == number of subjects (no missing phenotype)
    //Z_mat == matrix of covariates (including constant)
    //Z_mat.rows() == number of covariates (including constant)
    //The SNP will probably be included as a row in Z_mat 
     //(but it is ultimately optional as far as this method is
    //concerned - Z_mat should be set before calling this method).
    //Z_mat.cols() == number of subjects (no missing phenotype)
    //measure_vector : std::vector<measurement> used 
    //for omics measurement.
    //measure_vector.size() == number of subjects 
    //(no missing phenotype)
    //MAX_ITER : number of iterations allowed for the EM algorithm
    //CONVERGED_RATIO : double value passed from
    //this method to isVectorConverged()
    //post: All the non-constant reference arguments
    //will contain estimates.
    int perform_EM( double & gamma,
                    double & tau_sqr,
                    VectorXd & beta, VectorXd & ksi,
                    const VectorXd & Y_vec,
                    const MatrixXd & X_mat,
                    const MatrixXd & Z_mat,
                    const std::vector<measurement> & measure_vector,
                    VectorXd & se_vec,
                    size_t MAX_ITER,
                    double CONVERGED_RATIO,
                    MatrixXd & covariance_mat,
                    VectorXd & full_estimate_vec
                   );

    int calc_beta_gamma(
        const std::vector<double> & e_g1_vector,
        const std::vector<double> & e_g3_vector,
        const std::vector<double> & e_g4_vector,
        const std::vector<double> & e_g5_vector,
        const std::vector<double> & e_g6_vector,
        const std::vector<double> & e_g7_vector,
        const VectorXd & Y_vec,
        const MatrixXd & X_mat,
        const VectorXd & old_beta,
              VectorXd & new_beta,  
              double   old_gamma,
              double & new_gamma );

    
    int calc_ksi( const std::vector< double > & e_g1_vector,
                  const MatrixXd & Z_mat,
                        VectorXd & ksi );

    double calc_tau_sqr( 
        const std::vector< double > & e_g1_vector,
        const std::vector< double > & e_g2_vector,
        const VectorXd & ksi,
        const MatrixXd & Z_mat );
    
    void calc_Q_mat( 
        const std::vector< double > & e_g5_vector,
        const std::vector< double > & e_g6_vector,
        const std::vector< double > & e_g7_vector,
        const MatrixXd & X_mat,
        const MatrixXd & Z_mat,
        double tau_sqr,
              MatrixXd & Q_mat );                      

    void calc_V_mat_vec(
        double tau_sqr,
        const VectorXd & Y_vec,
        const VectorXd & ksi,
        const MatrixXd & X_mat,
        const MatrixXd & Z_mat,
        std::vector< MatrixXd >  & V_mat_vec );

    void calc_expect_U_sqr_mat_vec(
        const std::vector< double > & e_g1_vector,
        const std::vector< double > & e_g2_vector,
        const std::vector< double > & e_g3_vector,
        const std::vector< double > & e_g4_vector,
        const std::vector< double > & e_g8_vector,
        const std::vector< double > & e_g9_vector,
        const std::vector< double > & e_g10_vector,
        const std::vector< double > & e_g11_vector,
        const std::vector< double > & e_g12_vector,
        const std::vector< double > & e_g13_vector,
        const std::vector< double > & e_g14_vector,
        const std::vector< MatrixXd > & V_mat_vec,
              std::vector< MatrixXd > & expect_U_sqr_mat_vec );

    int calc_expect_U_vec_vec( 
        const std::vector< double > & e_g1_vector,
        const std::vector< double > & e_g2_vector,
        const std::vector< double > & e_g3_vector,
        const std::vector< double > & e_g4_vector,
        const std::vector< MatrixXd > & V_mat_vec,
              std::vector< VectorXd > & expect_U_vec_vec );


    void calc_omega_mat( 
        const std::vector<measurement> & measure_vector,
        const MatrixXd & Q_mat,
        const std::vector< MatrixXd > & expect_U_sqr_mat_vec,
        const std::vector< VectorXd > & expect_U_vec_vec,
              MatrixXd & omega_mat );

} //namespace ldl_logistic
#endif
